use core::f32;
use std::time::Duration;

use bevy::{
    asset::RenderAssetUsages,
    pbr::MaterialExtension,
    picking::PickSet,
    prelude::*,
    render::mesh::{Indices, PrimitiveTopology},
    tasks::AsyncComputeTaskPool,
    utils::HashMap,
};
use noise::{Fbm, MultiFractal, NoiseFn, Perlin};
use rand::{thread_rng, Rng, SeedableRng};
use voronoice::{BoundingBox, Point, VoronoiBuilder};

mod animals;
mod cells;
mod picking;
pub use animals::AnimalKind;
pub use cells::CellKind;
use cells::{
    material::{CellMaterial, CellMaterialExtension},
    *,
};
use picking::*;

use crate::map::cells::grow::{grow_thread, GrowThread};

pub struct Plugin;
impl bevy::prelude::Plugin for Plugin {
    fn build(&self, app: &mut App) {
        app.add_plugins(cells::Plugin)
            .add_systems(Startup, setup)
            .add_systems(PreUpdate, picking_backend.in_set(PickSet::Backend))
            .add_systems(Update, update_chunk_mesh)
            .insert_resource(ClearColor(Color::srgb(0., 0., 1.)))
            .insert_resource(Seed(thread_rng().gen()));
    }
}

/// Determined empirically
pub const AVERAGE_NEIGHBORS_NUMBER: f64 = 6.;

pub const HEIGHT: f64 = 4.;
pub const WIDTH: f64 = 4.;
pub const REAL_HEIGHT: f32 = 8000.;
pub const REAL_WIDTH: f32 = 8000.;
pub const CELL_AREA: f32 = REAL_HEIGHT * REAL_WIDTH / CELLS as f32;
pub const CELLS_TARGET_NUMBER: usize = 100000;
pub const CHUNKS_RESOLUTION: usize = 4;
pub const CHUNKS: usize = CHUNKS_RESOLUTION.pow(2);
pub const CELLS_PER_CHUNK: usize = CELLS_TARGET_NUMBER / CHUNKS;
pub const CELLS: usize = CELLS_PER_CHUNK * CHUNKS;

#[derive(Resource)]
struct Seed(u32);

#[derive(Resource)]
pub struct Voronoi(voronoice::Voronoi);

#[derive(Component)]
pub struct MeshMarker;

#[derive(Component)]
struct MeshColors(Vec<[f32; 4]>);

#[derive(Resource)]
pub struct CellsEntities(Vec<Entity>);

#[derive(Component)]
pub struct MeshNeedsUpdate(bool);

#[derive(Component)]
pub struct Chunk(usize);

fn setup(
    mut cmds: Commands,
    mut meshes: ResMut<Assets<Mesh>>,
    // mut materials: ResMut<Assets<StandardMaterial>>,
    mut materials: ResMut<Assets<CellMaterial>>,
    seed: Res<Seed>,
) {
    let mut rng = rand::rngs::SmallRng::seed_from_u64(seed.0 as u64);
    let mut sites = Vec::with_capacity(CELLS);
    for chunk_x in 0..CHUNKS_RESOLUTION {
        let min_x = WIDTH / 2. * ((2. * (chunk_x as f64) / (CHUNKS_RESOLUTION as f64)) - 1.);
        let max_x = min_x + (WIDTH as f64 / CHUNKS_RESOLUTION as f64);
        for chunk_y in 0..CHUNKS_RESOLUTION {
            let min_y = HEIGHT / 2. * ((2. * (chunk_y as f64) / (CHUNKS_RESOLUTION as f64)) - 1.);
            let max_y = min_y + (HEIGHT as f64 / CHUNKS_RESOLUTION as f64);
            for _ in 0..CELLS_PER_CHUNK {
                sites.push(Point {
                    x: rng.gen_range(min_x..max_x),
                    y: rng.gen_range(min_y..max_y),
                });
            }
        }
    }
    // dbg!(sites.len());

    let voronoi = VoronoiBuilder::default()
        .set_sites(sites)
        .set_bounding_box(BoundingBox::new_centered(WIDTH, HEIGHT))
        .set_lloyd_relaxation_iterations(3)
        .build()
        .unwrap();

    let mut cells = Vec::with_capacity(CELLS);
    let z_noise = Fbm::<Perlin>::new(seed.0);
    let moisture_noise = Fbm::<Perlin>::new(seed.0 + 1).set_frequency(2.);
    for i in 0..CELLS {
        let c = voronoi.cell(i);
        let site = c.site_position();
        let z = get_altitude(&z_noise, &[site.x as f32, site.y as f32]);
        let _m = (
            (moisture_noise.get([site.x, site.y]) + 1.) / 2.
            // Noise + [0; 1]
        )
        .clamp(0., 1.) as f32;
        let k = if z <= 0. {
            CellKind::Sea
        } else if z <= 0.04 {
            CellKind::Beach
        } else if z < 0.6 {
            CellKind::Dirt
        } else {
            CellKind::Stone
        };
        cells.push(Cell {
            kind: k,
            voronoi_id: i,
        });
    }

    // let mut poss = [Vec::new(); CHUNKS];
    let mut indices: [Vec<u32>; CHUNKS] = std::array::from_fn(|_| Vec::new());
    // let mut normals = Vec::new();

    let mut poss = voronoi.sites().chunks_exact(CELLS_PER_CHUNK).map(|ch| {
        ch.iter()
            .map(|pos| {
                let z = get_altitude(&z_noise, &[pos.x as f32, pos.y as f32]);
                Vec3::new(pos.x as f32, pos.y as f32, z)
            })
            .collect::<Vec<_>>()
    });
    let mut poss: [Vec<Vec3>; CHUNKS] = std::array::from_fn(|_| poss.next().unwrap());

    // for (i, pos) in voronoi.sites().iter().enumerate() {
    //     let z = get_altitude(&z_noise, &[pos.x as f32, pos.y as f32]);
    //     poss.push(Vec3::new(pos.x as f32, pos.y as f32, z));
    //     cells[i].vertices.push(i);
    // }
    // let mut hybrid_cells: HashMap<usize, usize> = HashMap::new();
    for t in voronoi.triangulation().triangles.chunks_exact(3) {
        let on_hull = t
            .iter()
            .filter(|v| voronoi.triangulation().hull.contains(v))
            .count();
        // Don't draw triangles on the hull that are often long and look glitchy
        if on_hull > 0 {
            continue;
        }
        let mut chs = t.iter().map(|c| Cell::chunk_from_id(*c));
        let chs: [usize; 3] = std::array::from_fn(|_| chs.next().unwrap());
        for ch in 0..CHUNKS {
            if !chs.contains(&ch) {
                continue;
            }
            // let ch = if chs[1] == chs[2] { chs[1] } else { chs[0] };
            // Add vertex to chunk if it is from an external chunk but we need it for a triangle. Else, just make the triangle
            if chs[2] != ch {
                poss[ch].push(poss[chs[2]][t[2] % CELLS_PER_CHUNK]);
                // hybrid_cells.insert(chs[2], chs[0]);
                indices[ch].push((poss[ch].len() - 1) as u32);
            } else {
                indices[ch].push((t[2] % CELLS_PER_CHUNK) as u32);
            }
            if chs[1] != ch {
                poss[ch].push(poss[chs[1]][t[1] % CELLS_PER_CHUNK]);
                // hybrid_cells.insert(chs[1], chs[0]);
                indices[ch].push((poss[ch].len() - 1) as u32);
            } else {
                indices[ch].push((t[1] % CELLS_PER_CHUNK) as u32);
            }
            if chs[0] != ch {
                poss[ch].push(poss[chs[0]][t[0] % CELLS_PER_CHUNK]);
                // hybrid_cells.insert(chs[0], chs[1]);
                indices[ch].push((poss[ch].len() - 1) as u32);
            } else {
                indices[ch].push((t[0] % CELLS_PER_CHUNK) as u32);
            }
        }
    }

    let mut cells_entities = Vec::with_capacity(cells.len());
    for (ch_id, ((poss, indices), ch_cells)) in poss
        .into_iter()
        .zip(indices.into_iter())
        .zip(cells.chunks_exact(CELLS_PER_CHUNK))
        .enumerate()
    {
        let colors = vec![[0.; 4]; poss.len()];
        let mut mesh = Mesh::new(
            PrimitiveTopology::TriangleList,
            RenderAssetUsages::default(),
        )
        // Add 4 vertices, each with its own position attribute (coordinate in
        // 3D space), for each of the corners of the parallelogram.
        .with_inserted_attribute(Mesh::ATTRIBUTE_POSITION, poss)
        .with_inserted_attribute(Mesh::ATTRIBUTE_COLOR, colors.clone())
        // .with_inserted_attribute(Mesh::ATTRIBUTE_NORMAL, normals)
        .with_inserted_indices(Indices::U32(indices));

        mesh.compute_smooth_normals();

        cmds.spawn((
            Mesh3d(meshes.add(mesh)),
            // StandardMaterial
            // MeshMaterial3d(materials.add(StandardMaterial::default())),
            MeshMaterial3d(materials.add(CellMaterial {
                base: StandardMaterial {
                    alpha_mode: AlphaMode::Mask(0.5),
                    ..Default::default()
                },
                extension: CellMaterialExtension {},
            })),
            Transform::default(),
            MeshColors(colors),
            MeshNeedsUpdate(true),
            MeshMarker,
            Chunk(ch_id),
        ))
        .with_children(|parent| {
            for cell in ch_cells {
                let kind = cell.kind;
                // let id = cell.voronoi_id;
                let mut cmd = parent.spawn(cell.clone());
                // if let Some(ch) = hybrid_cells.get(&id) {
                //     cmd.insert(GhostCell(*ch));
                // }
                cmd.observe(self::cells::input::on_click);
                cells_entities.push(cmd.id());
            }
        });
    }

    cmds.insert_resource(Voronoi(voronoi));
    cmds.insert_resource(CellsEntities(cells_entities));

    let task_pool = AsyncComputeTaskPool::get();
    cmds.insert_resource(GrowThread(task_pool.spawn(grow_thread())));
}

// TODO: update this to take chunks into account
fn update_chunk_mesh(
    cells: Query<&Cell>,
    mut chunks: Query<
        (&Mesh3d, &mut MeshColors, &mut MeshNeedsUpdate, &Children),
        With<MeshMarker>,
    >,
    mut meshes: ResMut<Assets<Mesh>>,
) {
    for (mesh, mut cols, mut needs_update, children) in chunks.iter_mut() {
        if needs_update.0 {
            if let Some(mesh) = meshes.get_mut(mesh) {
                let mut modified = false;
                // let mut rng = thread_rng();
                for child in children.iter() {
                    let cell = cells.get(*child).unwrap();
                    // let col: [f32; 4] = [rng.gen(), rng.gen(), rng.gen(), 1.];
                    let col = cell.color();
                    modified = modified || cols.0[cell.voronoi_id % CELLS_PER_CHUNK] != col;
                    cols.0[cell.voronoi_id % CELLS_PER_CHUNK] = col;
                    // for id in cell.vertices.iter() {
                    //     modified = modified || cols.0[*id] != col;
                    //     cols.0[*id] = col.clone();
                    // }
                }
                if modified {
                    mesh.insert_attribute(Mesh::ATTRIBUTE_COLOR, cols.0.clone());
                }
            }
            needs_update.0 = false;
            // Update at most one chunk per frame to avoid freezes
            return;
        }
    }
}

/// Between 0 and 1
fn get_altitude(z_noise: &Fbm<Perlin>, pos: &[f32; 2]) -> f32 {
    let z_noise = ((z_noise.get([pos[0] as f64, pos[1] as f64]) as f32) + 1.) / 2.; // Noise [0; 1]
    (
        -0.2// Arbitrary value
        + (z_noise.exp() / 1f32.exp())
        // Noise + [0; 1]
        // - ((pos[0].powi(2)+pos[1].powi(2)).sqrt()*0.3)
        // Distance - [0; sqrt(2)] * 0.5
    )
    // .clamp(-1., 1.)
}
