use std::time::Duration;

use bevy::{prelude::*, time::common_conditions::on_timer};
use rand::{seq::IteratorRandom, thread_rng, Rng};

use crate::{time::GameTime, ui::CurrentAction};

use super::{
    animals::Animal, AnimalKind, CellsEntities, MeshMarker, MeshNeedsUpdate, Voronoi,
    AVERAGE_NEIGHBORS_NUMBER, CELLS_PER_CHUNK, CELL_AREA,
};

pub mod grow;
pub mod input;
pub mod material;
use material::CellMaterial;

pub struct Plugin;
impl bevy::prelude::Plugin for Plugin {
    fn build(&self, app: &mut App) {
        // app.add_systems(
        //     FixedUpdate,
        //     (
        //         cells_regeneration.run_if(on_timer(Duration::from_millis(500))),
        //         expand.run_if(on_timer(Duration::from_millis(300))),
        //     ),
        // )
        app.add_plugins(MaterialPlugin::<CellMaterial>::default());
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum CellKind {
    Sea,
    Beach,
    Forest { wealth: WealthType },
    Dirt,
    Stone,
    Grass { wealth: WealthType },
}
impl CellKind {
    pub const fn regen_full_growth_duration(&self) -> Duration {
        match self {
            CellKind::Sea | CellKind::Beach | CellKind::Dirt | CellKind::Stone => unreachable!(),
            CellKind::Forest { .. } => Duration::from_secs(100 * 365 * 24 * 60 * 60), // Let's say that a forest takes 100 years to mature
            CellKind::Grass { .. } => Duration::from_secs(7 * 7 * 24 * 60 * 60), // Let's say that grass takes 7 weeks to reach its max
        }
    }
    pub const fn can_place_animal(&self, kind: AnimalKind) -> bool {
        match self {
            CellKind::Sea => false,
            _ => true,
        }
    }
    pub const fn grass_wealth(&self) -> Option<usize> {
        match self {
            CellKind::Grass { wealth } => Some(*wealth),
            CellKind::Forest { wealth } => Some(
                wealth.saturating_mul(
                    (CellKind::Forest { wealth: 0 }
                        .regen_full_growth_duration()
                        .as_secs()
                        / CellKind::Grass { wealth: 0 }
                            .regen_full_growth_duration()
                            .as_secs()) as usize,
                ),
            ),
            _ => None,
        }
    }
}

#[derive(Debug, Component, Clone)]
pub struct Cell {
    pub kind: CellKind,
    pub voronoi_id: usize,
}
impl Cell {
    pub const fn color(&self) -> [f32; 4] {
        match self.kind {
            CellKind::Sea => [0., 0., 1., 1.],
            CellKind::Beach => [0.82, 0.84, 0.51, 1.],
            CellKind::Forest { wealth } => [0., 0.5 - (wealth_to_unit(wealth) * 0.4), 0., 1.],
            CellKind::Dirt => [0.53, 0.38, 0.29, 1.],
            CellKind::Stone => [0.5, 0.5, 0.5, 1.],
            CellKind::Grass { wealth } => [
                (136. / 255.) - (wealth_to_unit(wealth) * 0.15),
                (154. / 255.) + (wealth_to_unit(wealth) * 0.1),
                (59. / 255.) - (wealth_to_unit(wealth) * 0.15),
                1.,
            ],
        }
    }
    pub const fn chunk_from_id(id: usize) -> usize {
        id / CELLS_PER_CHUNK
    }
    pub const fn chunk(&self) -> usize {
        Self::chunk_from_id(self.voronoi_id)
    }
}

// #[derive(Component)]
// /// Cell that's in another chunk but needs to be there to form a triangle
// pub struct GhostCell {
//     pub chunk:
// };

pub type WealthType = usize;
const fn wealth_to_unit(wealth: WealthType) -> f32 {
    wealth as f32 / WealthType::MAX as f32
}

// const STEP: WealthType = WealthType::MAX / 4;
// pub fn cells_regeneration(
//     mut cells: Query<(&mut Regeneration, &mut Wealth, &Parent)>,
//     mut chunks: Query<&mut MeshNeedsUpdate, With<MeshMarker>>,
//     gt: Res<GameTime>,
// ) {
//     for (mut regen, mut wealth, parent) in cells.iter_mut() {
//         while gt.current - regen.last_update > regen.full_growth_duration / u8::MAX as u32 {
//             regen.last_update = gt
//                 .current
//                 .min(regen.last_update + (regen.full_growth_duration / u8::MAX as u32));
//             wealth.0 = wealth.0.saturating_add(1);
//             // Not update map each time for optimization
//             if wealth.0 % STEP == 0 {
//                 chunks.get_mut(parent.get()).unwrap().0 = true;
//             }
//         }
//     }
// }

// pub fn expand(
//     mut cells: Query<(&mut Cell, Option<&Wealth>)>,
//     voronoi: Res<Voronoi>,
//     cells_entities: Res<CellsEntities>,
//     mut cmds: Commands,
//     t: Res<Time>,
//     gt: Res<GameTime>,
// ) {
//     let mut random = thread_rng();
//     let mut changes = Vec::new();
//     for (cell, wealth) in cells.iter() {
//         // TODO: search beter numbers, for now they are arbitrary
//         if cell.kind == CellKind::Forest || cell.kind == CellKind::Grass {
//             let wealth = if cell.kind == CellKind::Forest {
//                 1.
//             } else {
//                 wealth.unwrap().0 as f64 / 255.
//             };
//             if (t.elapsed_secs_f64() * (gt.speed as f64) * wealth * 1e-7
//                 / (CELL_AREA as f64).sqrt())
//                 >= 1.
//             {
//                 // "Grass can expand into adjacent dirt at a rate of 2.5 to 7.5 cm (1 to 3 inches) per month.", DeepSeek
//                 // With some (way too complicated and probably false) computation, we get for a good grass an extension rate of (1e-7 / sqrt(cell_area)) cells per seconds
//                 let target = voronoi
//                     .0
//                     .cell(cell.voronoi_id)
//                     .iter_neighbors()
//                     .choose(&mut random)
//                     .unwrap();
//                 changes.extend(
//                     voronoi
//                         .0
//                         .cell(cell.voronoi_id)
//                         .iter_neighbors()
//                         .map(|t| (t, CellKind::Grass)),
//                 );
//             }
//         }
//         if cell.kind == CellKind::Forest {
//             if (t.elapsed_secs_f64()
//                 * (gt.speed as f64)
//                 * (wealth.unwrap().0 as f64 / 255.).sqrt()
//                 * 5.6e-7
//                 / (CELL_AREA as f64).sqrt()
//                 * AVERAGE_NEIGHBORS_NUMBER)
//                 >= 1.
//             {
//                 // "Forests can expand at a rate of 1 to 10 meters per year", DeepSeek
//                 // Same computations as above : (5.6e-7 / sqrt(cell_area)) cells per seconds
//                 let target = voronoi
//                     .0
//                     .cell(cell.voronoi_id)
//                     .iter_neighbors()
//                     .choose(&mut random)
//                     .unwrap();
//                 changes.push((target, CellKind::Forest));
//             }
//         }
//     }
//     for (v_id, kind) in changes {
//         let target = cells_entities.0[v_id];
//         let (mut cell, _) = cells.get_mut(target).unwrap();
//         if kind == CellKind::Forest && (cell.kind == CellKind::Dirt || cell.kind == CellKind::Grass)
//         {
//             cell.kind = CellKind::Forest;
//             cmds.entity(target).insert((
//                 Wealth(0),
//                 Regeneration {
//                     full_growth_duration: CellKind::Forest.regen_full_growth_duration(),
//                     last_update: gt.current,
//                 },
//             ));
//         } else if kind == CellKind::Grass && cell.kind == CellKind::Dirt {
//             cell.kind = CellKind::Grass;
//             cmds.entity(target).insert((
//                 Wealth(0),
//                 Regeneration {
//                     full_growth_duration: CellKind::Grass.regen_full_growth_duration(),
//                     last_update: gt.current,
//                 },
//             ));
//         }
//     }
// }
